<?php

/**
 * @file
 *  IQ Extras - Incubating new Drush Issue Queue features
 *
 *  iq-submit- submit a patch for an issue
 *  iq-open - quickly open one of the files modified by the last patch
 */

/**
 * Implementation of hook_drush_command().
 */
function iq_extras_drush_command() {
  $items['iq-submit'] = array(
    'description' => 'Submit a patch for an issue on drupal.org.',
    'examples' => array(
      'drush iq-submit "This is my new patch"' => 'Make a patch from the current branch. The branch must have been created by drush iq-apply-patch.',
      'drush iq-submit "This is my patch" 123456' => 'Make a patch from the current branch and/or unstaged changes, and submit it to the specified issue id.',
      'drush iq-submit 123456 "This is my patch"' => 'Note that the issue id can come before or after the commit comment.',
    ),
    'arguments' => array(
      'comment' => 'Description of the patch.',
      'issue' => 'Issue id to submit the patch to.',
    ),
    'options' => array(
      'user' => 'The username to submit the patch as.',
      'pass' => 'The password to use to authenticate the username with.',
      'interdiff' => array('description' => 'Select the size ratio at which an interdiff will be made. Default is 0.3.',
        'example-value' => '0.3',
      ),
      'no-interdiff' => 'Prevent interdiff from being created. Equivalent to --interdiff=0',

      'title' => 'Set the issue title.',
      'version' => 'Set the issue version.',
      'component' => 'Set the issue component.',
      'assigned' => 'Set the drupal.org user who the issue is assigned to.',
      'category' => 'Set the issue category.',
      'priority' => 'Set the issue priority.',
      'status' => 'Set the issue status.',
      'tags' => 'Set the issue tags.  Note that this overwrites any existing tags.',
      'nw' => 'Needs work.  By default, iq-submit will set the status of an issue to "Needs review" unless --status is used to set it to something else.  --nw is shorthand for --status="needs work".'
    ),
    'aliases' => array('iqs'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['iq-open'] = array(
    'description' => 'Prompt the user and then open one file that has changed on the current branch.',
    'examples' => array(
      'drush iq-open' => 'Open one of the files changed by the last iq-apply-patch.',
      'drush iq-open --search="modification"' => 'Open one of the files changed by the last iq-apply-patch, where "modification" was added or removed by that patch.'
    ),
    'arguments' => array(
      'filter' => 'A substring for filtering the list of files. Omit this argument to choose from all patched files.',
    ),
    'options' => array(
      'all' => 'Open all changed files without prompting.',
      'grep' => 'Only consider files where the specified regular expression was added or removed by the last applied patch.',
      'search' => 'Only consider files where the specified search string was added or removed by the last applied patch.',
    ),
    'aliases' => array('iqo'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'topics' => array('docs-iq-commands'),
  );
  $items['docs-iq-extras'] = array(
    'description' => 'Iq-open and iq-submit documentation.',
    'hidden' => TRUE,
    'topic' => TRUE,
    'bootstrap' => DRUSH_BOOTSTRAP_DRUSH,
    'callback' => 'drush_print_file',
    'callback arguments' => array(dirname(__FILE__) . '/README.txt'),
  );

  return $items;
}

/**
 * Command hook for drush iq-open
 */
function drush_iq_extras_iq_open($filter = NULL) {
  // "simulated" mode's default implementation does not work
  // well here, so we'll turn it off, and handle it manually.
  // The problem is that, by default, in simulated mode,
  // calls to Drush's wrappers for 'exec', 'system', etc.
  // do not execute.  We make use of the results from a number
  // of calls to 'exec' to find the candidate files, so
  // it is easier to turn off simulated mode and continue
  // to use the wrappers. We will also set DRUSH_VERBOSE, as
  // this will cause our shell commands to be echo'ed as
  // they are executed.
  $simulated = drush_get_context('DRUSH_SIMULATE', FALSE);
  drush_set_context('DRUSH_SIMULATE', FALSE);
  if ($simulated) {
    drush_set_context('DRUSH_VERBOSE', TRUE);
  }
  // The first step is to use some helpful functions
  // from drush_iq to find out what our upstream branch
  // is.  If iq-apply-patch created the branch, then
  // $branch_merges_with will contain the branch that
  // the patch was applied to -- e.g. 7.x-1.x.
  // iq-open can also be used to select a file to open
  // from the current unstaged changes.  In this case,
  // $branch_merges_with will come back either empty
  // (no remote), or will be equal to $branch (if
  // our upstream is a remote branch of the same name).
  // In this case, we will use an empty branch spec
  // to just get a list of unstaged modified files.
  $branch_spec = '';
  $dir = _drush_iq_git_root_dir_from_cwd();
  $branch = _drush_iq_get_branch_at_dir($dir);
  $branch_merges_with = _drush_iq_get_branch_merges_with($dir);
  if (!empty($branch_merges_with) && ($branch != $branch_merges_with)) {
    $branch_spec = "${branch_merges_with}..HEAD";
  }
  // If the user has specified a --search (literal)
  // or --grep (regex) search term, then we will add
  // a -S or -G search option onto our git command to
  // filter the results.
  $selection = FALSE;
  $search_spec = '';
  foreach (array('S' => 'search', 'G' => 'grep') as $git_option => $drush_option) {
    $value = drush_get_option($drush_option, FALSE);
    if ($value) {
      $search_spec = "-$git_option" . drush_escapeshellarg($value);
    }
  }
  // git diff --name-only will list just the filenames that
  // were modified.
  $result = drush_shell_exec("git diff --name-only $search_spec %s", $branch_spec);
  $changed_files = drush_shell_exec_output();
  if (empty($changed_files) && empty($search_spec)) {
    drush_log(dt("No modified files."), 'warning');
    return;
  }
  // The user can further filter the selected files
  // by filename.  If you know that a modified file
  // contains "foo" in its name, then `drush iqo foo`
  // will open the file you want without requiring
  // you to type the full path.
  if ($filter) {
    foreach ($changed_files as $key => $file) {
      if (strpos($file, $filter) === FALSE) {
        unset($changed_files[$key]);
      }
    }
  }
  if (empty($changed_files)) {
    drush_log(dt("No matching files."), 'warning');
    return;
  }
  // drush_get_editor will return the editor to
  // use; this is often specified via the EDITOR
  // or VISUAL environment variables.  The default
  // is 'vi'.
  $exec = drush_get_editor();
  // Figure out which file or files to open, or
  // ask the user to pick one.
  if (drush_get_option('all', FALSE)) {
    $selection = implode(' ', $changed_files);
  }
  elseif (count($changed_files) == 1) {
    $selection = array_pop($changed_files);
  }
  else {
    $choice = drush_choice($changed_files, dt("Select a file to open"));
    if ($choice) {
      $selection = $changed_files[$choice];
    }
  }
  // Restore handling of simulated mode.
  drush_set_context('DRUSH_SIMULATE', $simulated);
  if ($selection) {
    drush_shell_exec_interactive($exec, $selection, $selection);
  }
}

/**
 * Command hook for drush iq-submit
 */
function drush_iq_extras_iq_submit($comment = '', $number = '') {
  // Handle "simulated" mode as we do for iq-open, above.
  $simulated = drush_get_context('DRUSH_SIMULATE', FALSE);
  drush_set_context('DRUSH_SIMULATE', FALSE);
  if ($simulated) {
    drush_set_context('DRUSH_VERBOSE', TRUE);
  }
  // Issue numbers never have spaces; comments always should have.
  // If the parameters provided has a space in the issue number,
  // but no space in the comment, then swap them.
  if (!empty($number) && !empty($comment)) {
    if ((strpos($number, ' ') !== FALSE) && (strpos($comment, ' ') === FALSE)) {
      $move_comment_to_number = $comment;
      $comment = $number;
      $number = $move_comment_to_number;
    }
  }

  // Provide a default comment if the user did not provide one.
  if (empty($comment)) {
    $comment = dt("Patch submitted by the Drush iq-submit command.");
  }
  // If an issue number was given, then use it in preference
  // to the issue number saved in the branch name (if any).
  if (!empty($number)) {
    $issue_info = drush_iq_get_info($number);
  }
  else {
    $dir = _drush_iq_git_root_dir_from_cwd();
    $branch = _drush_iq_get_branch_at_dir($dir);
    $issue_info = drush_iq_get_info($branch);
  }
  if (empty($issue_info)) {
    return FALSE;
  }
  // Record the issue id that we will be submitting our
  // patch to.
  $id = $issue_info['id'];
  $patch_filename = $issue_info['projectName'] . '-' . $id . '.patch';

  // Start off using the 'drush iq-diff' command
  // to take a full diff of the current branch
  // against its upstream branch
  $values = drush_invoke_process('@none', 'iq-diff', array(), array(), array('integrate' => FALSE));
  $primary_patch = $values['output'];

  // We will use 'git diff' to make an interdiff.
  // If this is empty, or if the interdiff is close
  // in size to the full diff, then we will omit it
  // when we post the patch.
  $result = drush_shell_exec("git diff");
  $git_diff_output = drush_shell_exec_output();
  $interdiff = implode("\n", $git_diff_output);

  // Nothing to see here?  Stop if there are no code differences.
  if (empty($primary_patch) && empty($interdiff)) {
    return drush_set_error('DRUSH_IQ_SUBMIT_NO_DIFF', dt("Cannot submit a patch; there are no code differences."));
  }

  // If the primary patch is empty, but the interdiff
  // has content, then we'll use the output from
  // git diff as the primary patch.
  if (empty($primary_patch) && !empty($interdiff)) {
    $primary_patch = $interdiff;
    $interdiff = '';
  }

  // If the user specified that there should not be
  // an interdiff, then empty it out here.
  if (drush_get_option('no-interdiff', FALSE)) {
    $interdiff = '';
  }

  // If there is an interdiff, check to see if it
  // is large compared to the size of the primary patch.
  if (!empty($interdiff)) {
    $interdiff_size_ratio = drush_get_option('interdiff', '0.3');
    $patch_length = strlen($primary_patch);
    $interdiff_length = strlen($interdiff);
    $max_interdiff_lengh = $patch_length * $interdiff_size_ratio;
    drush_log(dt("Patch len=%plen, Interdiff len = %ilen. Ratio=%ratio & max=%max", array('%plen' => $patch_length, '%ilen' => $interdiff_length, '%ratio' => $interdiff_size_ratio, '%max' => $max_interdiff_lengh)), 'notice');
    if ($interdiff_length > $max_interdiff_lengh) {
      $interdiff = '';
    }
  }

  // Get the username and password to use when submitting the patch
  $drupal_org_user = drush_get_option('user');
  $drupal_org_pass = drush_get_option('pass');
  if (empty($drupal_org_user)) {
    return drush_set_error('DRUSH_IQ_SUBMIT_NO_USER', dt("No username specified; cannot submit patch."));
  }
  if (empty($drupal_org_pass)) {
    return drush_set_error('DRUSH_IQ_SUBMIT_NO_PASS', dt("No password specified; cannot submit patch."));
  }

  // Log on as the selected user
  $session_cookies = drush_iq_submit_user_login($drupal_org_user, $drupal_org_pass);
  if (empty($session_cookies)) {
    return drush_set_error('DRUSH_IQ_SUBMIT_LOGIN_FAILED', dt("Login failed; check username and password."));
  }
  drush_log(dt("Logged in to drupal.org as %name", array('%name' => $drupal_org_user)), 'ok');

  // Fetch the comment form for the specified issue number.
  $post_headers[] = 'User-Agent: Drush (+http://drush.org/)';
  $cookie_header = '';
  foreach ($session_cookies as $cookie_name => $cookie_elements) {
    $cookie_value = $cookie_elements['value'];
    $cookie_header .= " {$cookie_name}={$cookie_value};";
  }
  if (!empty($cookie_header)) {
    $post_headers[] = "Cookie: $cookie_header";
  }

  $post = curl_init("http://drupal.org/node/$id");
  curl_setopt($post, CURLOPT_FOLLOWLOCATION, TRUE);
  curl_setopt($post, CURLOPT_HTTPHEADER, $post_headers);
  curl_setopt($post, CURLOPT_HEADER, TRUE);
  curl_setopt($post, CURLOPT_RETURNTRANSFER, TRUE);
  curl_setopt($post, CURLOPT_ENCODING, '');
  curl_setopt($post, CURLOPT_SSL_VERIFYPEER, 0);
  $result = curl_exec($post);

  // $comment_fields will contain all of the names
  // and values of the form fields from the previous
  // POST.  We will use these values when submitting
  // the next form.
  $form = _drush_iq_submit_get_form_fields($result, 'comment-form');
  $comment_fields = $form['fields'];

  // Create a temporary directory to save our patch
  // file and interdiff in.
  $iq_submit_tmp_dir = drush_tempdir();
  $file_upload_list = array();

  // Write our primary patch, and if applicable, the
  // interdiff, to the temporary directory.
  file_put_contents($iq_submit_tmp_dir . '/' . $patch_filename, $primary_patch);
  $file_upload_list[] = $iq_submit_tmp_dir . '/' . $patch_filename;
  if ($interdiff) {
    file_put_contents($iq_submit_tmp_dir . '/interdiff.txt', $interdiff);
    $file_upload_list[] = $iq_submit_tmp_dir . '/interdiff.txt';
  }

  // Uploading the patch and interdiff is handled
  // identically; only the path to the file to upload
  // differs. We therefore use a short loop to upload them.
  foreach ($file_upload_list as $upload_file) {
    $upload_fields = $comment_fields;
    $upload_fields['attach'] = 'Attach';
    $upload_fields['files[upload]'] = '@' . $upload_file;
    unset($upload_fields['op']);

    $post_upload = curl_init("http://drupal.org" . $form['action']);
    curl_setopt($post_upload, CURLOPT_FOLLOWLOCATION, TRUE);
    curl_setopt($post_upload, CURLOPT_HTTPHEADER, $post_headers + array("Content-type: multipart/form-data"));
    curl_setopt($post_upload, CURLOPT_HEADER, TRUE);
    curl_setopt($post_upload, CURLOPT_RETURNTRANSFER, TRUE);
    curl_setopt($post_upload, CURLOPT_ENCODING, '');
    curl_setopt($post_upload, CURLOPT_SSL_VERIFYPEER, 0);
    curl_setopt($post_upload, CURLOPT_POST, 1);
    curl_setopt($post_upload, CURLOPT_POSTFIELDS, $upload_fields);
    $result = curl_exec($post_upload);
    if (curl_error($post_upload)) {
      print("Post error errno = " . curl_errno($post_upload));
    }
    // Note that once we POST our upload files, we once
    // again pull the latest form fields out of the page
    // text that is returned to us, so that we can use the
    // most recent values on the next form that we submit.
    $form = _drush_iq_submit_get_form_fields($result, 'comment-form');
    $comment_fields = $form['fields'];
    foreach($comment_fields as $key => $value) {
      if (substr($key, -8) == '[remove]') {
        unset($comment_fields[$key]);
      }
    }
    drush_log(dt("Attached %file.", array("%file" => basename($upload_file))), 'ok');
  }

  // Save the issue with its file uploads
  $comment_fields['op'] = 'Save';
  unset($comment_fields['attach']);
  $comment_fields['comment'] = $comment;

  // By default, we set the status to "needs review"
  // unless the user specifies that it should be something else.
  $defaults = array();
  $defaults['status'] = drush_get_option('nw', FALSE) ? "needs work" : "needs review";

  // Mapping table from the Drush command line option
  // name to the corresponding form field item to set.
  $option_to_form_item_name = array(
    'title' => 'title',
    'version' => 'project_info[rid]',
    'component' => 'project_info[component]',
    'assigned' => 'project_info[assigned]',
    'category' => 'category',
    'priority' => 'priority',
    'status' => 'sid',
    'tags' => 'taxonomy[tags][9]',
  );
  foreach ($option_to_form_item_name as $option_name => $form_item_name) {
    // If the user specified an issue metadata item to
    // change (e.g. --status='needs work'), then we will
    // get its value here. If the option is not set, then
    // drush_get_option will return the default value we
    // pass in as the second argument.
    $new_value = drush_get_option($option_name, isset($defaults[$option_name]) ? $defaults[$option_name] : FALSE);
    // If the option was specified, then change the form
    // value as the user desired.
    if ($new_value !== FALSE) {
      // If the form field is a SELECT element, then the
      // values that it may be set to will be found in
      // the form's 'possible-values' item.  In this case,
      // we map from the value specified by the user (which
      // is the label of the OPTION) with the corresponding
      // item's value.
      if (isset($form['possible-values'][$form_item_name])) {
        if (isset($form['possible-values'][$form_item_name][$new_value])) {
          $new_value = $form['possible-values'][$form_item_name][$new_value];
        }
        else {
          // If the user picked something not allowed, show
          // all of the valid choices.
          $valid_values = implode("\n  ", array_keys($form['possible-values'][$form_item_name]));
          return drush_set_error('DRUSH_IQ_SUBMIT_INVALID_VALUE', dt('"%name" cannot be given the value "%value".  Please select one of the valid options:' . "\n  %valid", array("%value" => $new_value, "%name" => $option_name, "%valid" => $valid_values)));
        }
      }
      $comment_fields[$form_item_name] = $new_value;
    }
  }

  $post_data = array();
  foreach ($comment_fields as $key => $value) {
    $post_data[] = $key . '=' . urlencode($value);
  }

  // Abort now if we are not going to submit this patch.
  if ($simulated) {
    drush_log(dt("Simulating an iq-submit; field values to post:\n%fields", array("%fields" => var_export($comment_fields, TRUE))), 'ok');
    return;
  }

  // Final step: save the prepared data.
  $post2 = curl_init("http://drupal.org" . $form['action']);
  curl_setopt($post2, CURLOPT_FOLLOWLOCATION, TRUE);
  curl_setopt($post2, CURLOPT_HTTPHEADER, $post_headers);
  curl_setopt($post2, CURLOPT_HEADER, TRUE);
  curl_setopt($post2, CURLOPT_RETURNTRANSFER, TRUE);
  curl_setopt($post2, CURLOPT_ENCODING, '');
  curl_setopt($post2, CURLOPT_SSL_VERIFYPEER, 0);
  curl_setopt($post2, CURLOPT_POST, count($post_data));
  curl_setopt($post2, CURLOPT_POSTFIELDS, implode('&', $post_data));
  $result = curl_exec($post2);
  drush_log(dt("Patch submitted.", 'success'));
}

/**
 * Log in to drupal.org as the specified user, with the password provided.
 */
function drush_iq_submit_user_login($user, $pass) {
  $session_cookies = array();
  // The first step to logging the user in is to fetch
  // the /user/login page, and retreive all of the field
  // item names and values from the user-login form.
  $contents = file_get_contents("https://drupal.org/user/login");
  $form = _drush_iq_submit_get_form_fields($contents, "user-login");
  $input_fields = $form['fields'];
  // Change the values of the user name and password to
  // whatever was provided by the user.
  $input_fields['name'] = $user;
  $input_fields['pass'] = $pass;
  $post_headers[] = 'User-Agent: Drush (+http://drush.org/)';

  $post_data = array();
  foreach ($input_fields as $key => $value) {
    $post_data[] = $key . '=' . urlencode($value);
  }

  $url = "https://drupal.org/user/login";
  $post = curl_init($url);
  curl_setopt($post, CURLOPT_FOLLOWLOCATION, TRUE);
  curl_setopt($post, CURLOPT_HTTPHEADER, $post_headers);
  curl_setopt($post, CURLOPT_HEADER, TRUE);
  curl_setopt($post, CURLOPT_RETURNTRANSFER, TRUE);
  curl_setopt($post, CURLOPT_ENCODING, '');
  curl_setopt($post, CURLOPT_SSL_VERIFYPEER, 0);
  curl_setopt($post, CURLOPT_POST, count($post_data));
  curl_setopt($post, CURLOPT_POSTFIELDS, implode('&', $post_data));
  $result = curl_exec($post);

  if (!curl_error($post)) {
    // Extract the headers from the result.  The regex
    // below deletes everything after the first blank line.
    $header_text = preg_replace('#^\s*$.*#smx', '', $result);
    $headers = _http_parse_headers($header_text);
    // Find the 'Set-Cookie' header.  It is necessary that
    // we stash and re-use the cookies provided to us so that
    // the session that drupal.org creates for us can be maintained.
    if (isset($headers['Set-Cookie'])) {
      foreach ($headers['Set-Cookie'] as $cookie) {
        list($cookie_name, $cookie_data) = explode('=', $cookie, 2);
        $cookie_elements = explode(';', $cookie_data);
        $session_cookies[$cookie_name]['value'] = array_shift($cookie_elements);
        foreach ($cookie_elements as $element) {
          $element = trim($element);
          if (strpos($element, '=') === FALSE) {
            $element_name = $element;
            $element_value = TRUE;
          }
          else {
            list($element_name, $element_value) = explode('=', $element, 2);
          }
          $session_cookies[$cookie_name][$element_name] = $element_value;
        }
      }
    }
  }
  curl_close($post);

  return $session_cookies;
}

/**
 * Extract the input, select and textarea elements
 * from the specified form.
 */
function _drush_iq_submit_get_form_fields($contents, $form_id) {
  $input_fields = array();
  $submit_buttons = array();
  $possible_values = array();

  // Clean up the form a bit before we parse it.
  // First delete everything before the <form ...>
  $contents = preg_replace('#.*(<form[^>]* id="' . $form_id . '"[^>]*>)#ms', '${1}', $contents);
  // Next, delete everything after the </form>
  $contents = preg_replace("#</form>.*#ms", "</form>", $contents);
  // Finally, we get better results from SimpleXMLElement
  // if we convert every <option ...>Label</option> element
  // into <option ... label="Label" />.
  $contents = preg_replace('#(<option[^>]*)>([^<]*)</option>#ms', '${1} label="${2}" />', $contents);

  $xml_form = new SimpleXMLElement("<?xml version='1.0' standalone='yes'?>" . $contents);
  foreach ($xml_form->xpath('//input') as $input_field) {
    if (!empty($input_field['name'])) {
      // Keep a separate list of all of the <submit ...> elements.
      if ($input_field['type'] == 'submit') {
        $submit_buttons[(string)$input_field['name']][] = (string)$input_field['value'];
      }
      else {
        $input_fields[(string)$input_field['name']] = (string)$input_field['value'];
      }
      // If the item is a checkbox, set its value to FALSE
      // if it is not checked.
      if (($input_field['type'] == 'checkbox') && (empty($input_field['checked']))) {
        $input_fields[(string)$input_field['name']] = FALSE;
      }
    }
  }
  foreach ($xml_form->xpath('//select') as $select_field) {
    if (!empty($select_field['name'])) {
      $selected = $select_field->xpath("option[@selected='selected']");
      // Record the name and value of the select item
      $input_fields[(string)$select_field['name']] = (string)$selected[0]['value'];
      // Map from 'label' to 'value'.  We will use
      // the 'possible-values' list to do field validation
      // if the user changes any of the issue's metadata
      // items with a commandline option.
      foreach ($select_field->xpath("option") as $option) {
        $possible_values[(string)$select_field['name']][(string)$option['label']] = (string)$option['value'];
      }
    }
  }
  foreach ($xml_form->xpath('//textarea') as $textarea) {
    // <textarea ...> elements are handled like <input ...> elements.
    $input_fields[(string)$textarea['name']] = (string)$textarea['value'];
  }
  $form['action'] = (string)$xml_form['action'];
  $form['submit'] = $submit_buttons;
  $form['fields'] = $input_fields;
  $form['possible-values'] = $possible_values;

  return $form;
}

// http_parse_headers is not always available, so we define
// our own copy of the function here.
//
// c.f. http://php.net/manual/en/function.http-parse-headers.php
function _http_parse_headers( $header )
{
  $retVal = array();
  $fields = explode("\r\n", preg_replace('/\x0D\x0A[\x09\x20]+/', ' ', $header));
  foreach( $fields as $field ) {
    if( preg_match('/([^:]+): (.+)/m', $field, $match) ) {
      $match[1] = preg_replace('/(?<=^|[\x09\x20\x2D])./e', 'strtoupper("\0")', strtolower(trim($match[1])));
      if( isset($retVal[$match[1]]) ) {
        if (!is_array($retVal[$match[1]])) {
          $retVal[$match[1]] = array($retVal[$match[1]]);
        }
        $retVal[$match[1]][] = $match[2];
      } else {
        $retVal[$match[1]] = trim($match[2]);
      }
    }
  }
  return $retVal;
}
